//
// Lighthouse3D.com
//
// Uses:
//  Devil for image loading
//		http://openil.sourceforge.net/
//
// The texture was downloaded from:
// http://opengameart.org/content/basket-ball-texture
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//


#include <cmath>
#include <ctime>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Include DevIL for image loading
#include <IL/il.h>

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"
#include "vsLogLib.h"

//----------------------------------------------------------------
// TODO - FIXED:
// 1) See how star positions and colors are set in function "init".
// 2) See how stars are rendered in function "renderScene".
// 3) Notice the texture load and blending in function "init".
// 4) Why is there an visual error in the blending?
// 5) Blend mode GL_ONE_MINUS_SRC_ALPHA vs GL_ONE, explain the difference.

// TODO - ITERATIVE:
// 1) Examine the changes in functions "setupShaders", "init" and "renderScene". Lot of CPU code!
// 2) Decrease the number of particles to better see what is happening.
//----------------------------------------------------------------

#define _FIXED		0
#define _ITERATIVE	1

#define _PARTICLES _FIXED
//----------------------------------------------------------------

VSMathLib *vsml;
VSShaderLib shader, shaderUpdate;

VSLogLib mLogError, mLogInfo;

// Vao
GLuint vao;

// Texture
unsigned int texID;

#if (_PARTICLES == _ITERATIVE)
// Double buffered position texture
GLuint posTexture[2];

/// Double buffered texture framebuffer
GLuint posFramebuffer[2];

int flip = 0;
#endif

// Camera Position
float camX, camY, camZ;

// Mouse Tracking Variables
int startX, startY, tracking = 0;

// Camera Spherical Coordinates
float alpha = 0.0f, beta = 0.0f;
float r = 1.0f;

// Geometry offset
float offsetX = 0.f, offsetY = 0.f;

// Grid state
const int PARTICLES_SQRT = 64;
float timeC = 0.f;

unsigned int loadRGBATexture(std::string filename, bool mipmap = true, bool compress = false, GLenum aFilter = GL_LINEAR, GLenum aRepMode = GL_REPEAT);

// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	float ratio;

	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	// Set the viewport to be the entire window
	glViewport(0, 0, w, h);

	ratio = (1.0f * w) / h;
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
}


// ------------------------------------------------------------
//
// Render stuff
//

void renderScene(void)
{
#if (_PARTICLES == _FIXED)
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);
	
	// Set camera
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);

	vsml->translate(offsetX, offsetY, 0.f);    

	// Use our shader
	glUseProgram(shader.getProgramIndex());

	// Send matrices to uniform buffer
	vsml->matricesToGL();

	shader.setUniform("time", timeC);

	// Render VAO
	glBindVertexArray(vao);
	glDrawArrays(GL_POINTS, 0, PARTICLES_SQRT * PARTICLES_SQRT);
#elif (_PARTICLES == _ITERATIVE)
	glUseProgram(shaderUpdate.getProgramIndex());

	// Bind to the writing buffer.
	glBindFramebuffer(GL_FRAMEBUFFER, posFramebuffer[(flip + 1) % 2]);

	// Save current viewport
	int viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	// Set the viewport to the texture size.
	glViewport(0, 0, PARTICLES_SQRT, PARTICLES_SQRT);

	// This texture contains the current particle data.
	glBindTexture(GL_TEXTURE_2D, posTexture[flip]);

	// Draw all particles, data is updated in the shader.
	glBindVertexArray(vao);
	glDrawArrays(GL_POINTS, 0, PARTICLES_SQRT * PARTICLES_SQRT);

	// ------------------------------------------------------------

	// Go back to the window buffer.
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	glBindTexture(GL_TEXTURE_2D, posTexture[flip]);

	// Reset the viewport.
	glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);

	// Switch to the next buffer.
	flip = (flip + 1) % 2;

	// ------------------------------------------------------------

	glClear(GL_COLOR_BUFFER_BIT);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);

	// Set camera
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);

	// Use our shader
	glUseProgram(shader.getProgramIndex());

	// Send matrices to uniform buffer
	vsml->matricesToGL();

	// Render VAO
	glBindVertexArray(vao);
	glEnable(GL_BLEND); // Blending must not be enabled during texture write!!!
	glDrawArrays(GL_POINTS, 0, PARTICLES_SQRT * PARTICLES_SQRT);
	glDisable(GL_BLEND);
#endif

	timeC += 0.01f;

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;

		case 'c': printf("Camera Spherical Coordinates (%f, %f, %f)\n", alpha, beta, r);
			break;

		case 'm': glEnable(GL_MULTISAMPLE);
			break;

		case 'n': glDisable(GL_MULTISAMPLE);
			 break;
	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//

void processMouseButtons(int button, int state, int xx, int yy)
{
	// Start tracking the mouse
	if (state == GLUT_DOWN)
	{
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
		else if (button == GLUT_MIDDLE_BUTTON)
			tracking = 3;
	}

	// Stop tracking the mouse
	else if (state == GLUT_UP)
	{
		if (tracking == 1)
		{
			alpha -= (xx - startX);
			beta += (yy - startY);
			if (beta > 85.0f)
				beta = 85.0f;
			else if (beta < -85.0f)
				beta = -85.0f;
		}
		else if (tracking == 2)
		{
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}

// Track mouse motion while buttons are pressed
void processMouseMotion(int xx, int yy)
{
	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// Left mouse button: move camera
	if (tracking == 1)
	{
		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// Right mouse button: zoom
	else if (tracking == 2)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}
	// Middle mouse button: move object
	else if (tracking == 3)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r;

		offsetX -= deltaX * 0.0001f;
		offsetY -= deltaY * 0.0001f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}

void mouseWheel(int wheel, int direction, int x, int y)
{
	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Shader for models
	shader.init();
#if (_PARTICLES == _FIXED)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/stars.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/stars.frag");
#elif (_PARTICLES == _ITERATIVE)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/particles.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/particles.frag");
#endif

	// Set semantics for the shader variables
	shader.setProgramOutput(0, "outputF");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
#if (_PARTICLES == _FIXED)
	shader.setVertexAttribName(VSShaderLib::VERTEX_ATTRIB1, "color");
#endif


	shader.prepareProgram();
	printf("InfoLog\n%s\n\n", shader.getAllInfoLogs().c_str());

	shader.setUniform("texUnit", 0); // Add sampler uniform

#if (_PARTICLES == _ITERATIVE)
	shader.setUniform("particleUnit", 1); // Set the particle texture for the particles.vert shader

	shaderUpdate.init();
	shaderUpdate.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/particle_update.vert");
	shaderUpdate.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/particle_update.frag");
	
	shaderUpdate.setProgramOutput(0, "outputF");
	shaderUpdate.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");

	shaderUpdate.prepareProgram();
	printf("InfoLog\n%s\n\n", shaderUpdate.getAllInfoLogs().c_str());

	shaderUpdate.setUniform("particleUnit", 1); // Set the particle texture for the particle_update.vert shader
#endif

	return(shader.isProgramLinked());
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

// helper function for derived classes
// loads an image and defines an 8-bit RGBA texture
unsigned int loadRGBATexture(std::string filename, 
						bool mipmap, bool compress, 
						GLenum aFilter, GLenum aRepMode) {

	ILboolean success;
	unsigned int imageID;
	GLuint textureID = 0;

	// Load Texture Map
	ilGenImages(1, &imageID); 
	
	ilBindImage(imageID); /* Binding of DevIL image name */
	ilEnable(IL_ORIGIN_SET);
	ilOriginFunc(IL_ORIGIN_LOWER_LEFT); 
	success = ilLoadImage((ILstring)filename.c_str());

	if (!success) {
		VSLOG(mLogError, "Couldn't load texture: %s", 
							filename.c_str());
		// The operation was not sucessfull 
		// hence free image and texture 
		ilDeleteImages(1, &imageID); 
		return 0;
	}

	// add information to the log

	VSLOG(mLogInfo, "Texture Loaded: %s", filename.c_str());
	printf("Width: %d, Height %d, Bytes per Pixel %d", 
				ilGetInteger(IL_IMAGE_WIDTH),
				ilGetInteger(IL_IMAGE_HEIGHT),
				ilGetInteger(IL_IMAGE_BYTES_PER_PIXEL));

	std::string s;
	switch(ilGetInteger(IL_IMAGE_FORMAT)) {
		case IL_COLOR_INDEX      : s =  "IL_COLOR_INDEX"; break;     
		case IL_ALPHA		     : s =  "IL_ALPHA"; break;	
		case IL_RGB              : s =  "IL_RGB"; break;     
		case IL_RGBA             : s =  "IL_RGBA"; break;     
		case IL_BGR              : s =  "IL_BGR"; break;     
		case IL_BGRA             : s =  "IL_BGRA"; break;     
		case IL_LUMINANCE        : s =  "IL_LUMINANCE"; break;     
		case IL_LUMINANCE_ALPHA  : s =  "IL_LUMINANCE_ALPHA"; break;
	}
	printf(" Format %s", s.c_str());
	
	switch(ilGetInteger(IL_IMAGE_TYPE)) {
		case IL_BYTE           : s =  "IL_BYTE"; break;     
		case IL_UNSIGNED_BYTE  : s =  "IL_UNSIGNED_BYTE"; break;	
		case IL_SHORT          : s =  "IL_SHORT"; break;     
		case IL_UNSIGNED_SHORT : s =  "IL_UNSIGNED_SHORT"; break;     
		case IL_INT            : s =  "IL_INT"; break;     
		case IL_UNSIGNED_INT   : s =  "IL_UNSIGNED_INT"; break;     
		case IL_FLOAT          : s =  "IL_FLOAT"; break;     
		case IL_DOUBLE         : s =  "IL_DOUBLE"; break;
		case IL_HALF           : s =  "IL_HALF"; break;
	}
	printf(" Data type:  %s\n", s.c_str());

	/* Convert image to RGBA */
	ilConvertImage(IL_RGBA, IL_UNSIGNED_BYTE); 
	// Set filters
	GLenum minFilter = aFilter;
	if (aFilter == GL_LINEAR && mipmap) {
		minFilter = GL_LINEAR_MIPMAP_LINEAR;
	}
	else if (aFilter == GL_NEAREST && mipmap){
		minFilter = GL_NEAREST_MIPMAP_LINEAR;
	}
	GLenum type;
	if (compress)
		type = GL_RGBA;
	else
		type = GL_COMPRESSED_RGBA;


	/* Create and load textures to OpenGL */
	glGenTextures(1, &textureID); /* Texture name generation */
	glBindTexture(GL_TEXTURE_2D, textureID); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, aFilter); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, aRepMode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, aRepMode);
	glTexImage2D(GL_TEXTURE_2D, 0, type, 
					ilGetInteger(IL_IMAGE_WIDTH),
					ilGetInteger(IL_IMAGE_HEIGHT), 
					0, GL_RGBA, GL_UNSIGNED_BYTE,
					ilGetData()); 

	// Mipmapping?
	if (mipmap)
		glGenerateMipmap(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D,0);

	/* Because we have already copied image data into texture data
	we can release memory used by image. */
	ilDeleteImages(1, &imageID); 


	return textureID;
}

float random(float low, float high)
{
	float r01 = ((float)rand() / (float)RAND_MAX);
	return low + r01 * (high - low);
}

bool init()
{
	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

	// Initialize random generator
	srand(time(NULL));

	// Create the VAO
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

#if (_PARTICLES == _FIXED)
	// Create buffers for our vertex data
	GLuint buffers[2];
	glGenBuffers(2, buffers);

	// Create vertex array - regular mesh of triangles
	const int particleCount = PARTICLES_SQRT*PARTICLES_SQRT;
	GLfloat *vertices = new float[particleCount*4];
	GLfloat *colors = new float[particleCount*3];

	GLfloat *vertex = vertices;
	GLfloat *color = colors;
	for(int i = 0; i < PARTICLES_SQRT; i++)
	{
		for(int j = 0; j < PARTICLES_SQRT; j++)
		{
			vertex[0] = random(-5.f, 5.f);
			vertex[1] = random(-5.f, 5.f);
			vertex[2] = random(-100.f, 0.f);
			vertex[3] = 1.f;
			vertex += 4; // Move to next vertex

			color[0] = random(0.f, 1.f);
			color[1] = random(0.f, 1.f);
			color[2] = random(0.f, 1.f);
			color += 3; // Move to next vertex
		}
	}

	// Vertex coordinates buffer
	glBindBuffer(GL_ARRAY_BUFFER, buffers[0]);
	glBufferData(GL_ARRAY_BUFFER, particleCount * 4 * sizeof(GLfloat), vertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(VSShaderLib::VERTEX_COORD_ATTRIB);
	glVertexAttribPointer(VSShaderLib::VERTEX_COORD_ATTRIB, 4, GL_FLOAT, 0, 0, 0);

	// Color buffer
	glBindBuffer(GL_ARRAY_BUFFER, buffers[1]);
	glBufferData(GL_ARRAY_BUFFER, particleCount * 3 * sizeof(GLfloat), colors, GL_STATIC_DRAW);
	glEnableVertexAttribArray(VSShaderLib::VERTEX_ATTRIB1);
	glVertexAttribPointer(VSShaderLib::VERTEX_ATTRIB1, 3, GL_FLOAT, 0, 0, 0);

	// Cleanup
	delete[] vertices;
	delete[] colors;

	// Unbind the VAO
	glBindVertexArray(0);
#elif (_PARTICLES == _ITERATIVE)
	// Initialize particle data

	const int particleCount = PARTICLES_SQRT*PARTICLES_SQRT;
	// Points for accessing the texel, which contain the particle data.
	GLfloat *points = new GLfloat[particleCount*2];
	// Array containing the initial particles.
	GLfloat *particles = new GLfloat[particleCount*4];

	for(int y = 0; y < PARTICLES_SQRT; y++)
	{
		for(int x = 0; x < PARTICLES_SQRT; x++)
		{
			points[y * PARTICLES_SQRT * 2 + x * 2 + 0] = (GLfloat) x / (GLfloat) PARTICLES_SQRT;
			points[y * PARTICLES_SQRT * 2 + x * 2 + 1] = (GLfloat) y / (GLfloat) PARTICLES_SQRT;
		}
	}

	for(int y = 0; y < PARTICLES_SQRT; y++)
	{
		for(int x = 0; x < PARTICLES_SQRT; x++)
		{
			particles[y * PARTICLES_SQRT * 4 + x * 4 + 0] = 0.0f;
			particles[y * PARTICLES_SQRT * 4 + x * 4 + 1] = 0.0f;
			particles[y * PARTICLES_SQRT * 4 + x * 4 + 2] = 0.0f;
			particles[y * PARTICLES_SQRT * 4 + x * 4 + 3] = -1.0f; // w < 0.0 means a dead particle. So renewed in first frame.
		}
	}

	// Create VAOs for the points
	GLuint bufferPoints;
	glGenBuffers(1, &bufferPoints);
	glBindBuffer(GL_ARRAY_BUFFER, bufferPoints);
	glBufferData(GL_ARRAY_BUFFER, particleCount * 2 * sizeof(GLfloat), (GLfloat*) points, GL_STATIC_DRAW);
	glEnableVertexAttribArray(VSShaderLib::VERTEX_COORD_ATTRIB);
	glVertexAttribPointer(VSShaderLib::VERTEX_COORD_ATTRIB, 2, GL_FLOAT, 0, 0, 0);

	// Unbind the VAO
	glBindVertexArray(0);
#endif

	/* initialization of DevIL */
	ilInit();

	texID = loadRGBATexture("textures/star.gif", false);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texID);

#if (_PARTICLES == _ITERATIVE)
	// Create textures for particle data
	glGenTextures(2, posTexture);
	glActiveTexture(GL_TEXTURE1);

	for(int i = 0; i < 2; i++)
	{
		glBindTexture(GL_TEXTURE_2D, posTexture[i]);

		// Create a texture containing floats.
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, PARTICLES_SQRT, PARTICLES_SQRT, 0, GL_RGBA, GL_FLOAT, particles);

		// Important: No interpolation.
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	}

	// NOTICE: The texture will be bound during render
	glBindTexture(GL_TEXTURE_2D, 0);

	// Create framebuffers for particle textures
	glGenFramebuffers(2, posFramebuffer);

	for(int i = 0; i < 2; i++)
	{
		// Attach the texture.
		glBindFramebuffer(GL_FRAMEBUFFER, posFramebuffer[i]);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, posTexture[i], 0);
	}

	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// Cleanup
	delete[] points;
	delete[] particles;
#endif

	// Some GL settings
	//glEnable(GL_DEPTH_TEST);
	glDisable(GL_DEPTH_TEST); // To prevent particle overdrawing!!!
	glEnable(GL_MULTISAMPLE);

	// Enable blending
#if (_PARTICLES == _FIXED)
	glEnable(GL_BLEND);
#if 1
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#else
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
#endif
#elif (_PARTICLES == _ITERATIVE)
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
#endif

	glClearColor(0.f, 0.f, 0.f, 1.0f);

	return true;
}

void initVSL()
{
//	Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Particle Systems");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc(mouseWheel);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	if(!init())
		exit(1);

	initVSL();

//  GLUT main loop
	glutMainLoop();

	return(0);
}
